#!/usr/bin/perl -w 
# Journal & report module
#
# Copyright (C) 2007-2008 The Linux Foundation. All rights reserved.
#
# This program has been developed by ISP RAS for LF with some parts
# derived from lsb-autotest written by Jiri Dluhos <jdluhos@suse.cz>
# Copyright (C) 2005-2007 SuSE Linux Products GmbH
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

# Determines the directory the script is executed from,
# and stores it to the global 'script_directory' variable.

my $script_directory;

# Ensures that the specified path is absolute, prepending the current
# working directory to relative paths.
sub make_path_absolute
{
    my $input_path = shift;

    # Determine the current working directory.
    my $working_directory = `pwd`;
    chomp ($working_directory);

    my $output_path;
    if ($input_path =~ /^\//) {
        # If the path is absolute (starts with a slash), leave it as it is.
        $output_path = $input_path;
    } else {
        # If the path is relative (does not start with a slash), make it
        # absolute by prepending the current working directory.
        $output_path .= $working_directory . "/" . $input_path;
    }
    
    # Remove double //
    $output_path =~ s/\/\//\//g while ($output_path =~ /\/\//);

    # Remove "/."
    $output_path =~ s/\/\.\//\//g while ($output_path =~ /\/\.\//);
    $output_path =~ s/\/\.?$//g while ($output_path =~ /\/\.?$/);
    
    return $output_path;
}

sub find_script_directory
{
    $script_directory = $0;
    $script_directory =~ s/[^\/]*$//; # Basename

    $script_directory = make_path_absolute($script_directory);
}

sub BEGIN {
    $| = 1; # Immediate flush of all output

    # Append the script directory to the @INC to attach our own modules
    find_script_directory();
    push @INC, $script_directory;
}

use strict;

use IO::File;
use Common qw( sendmail ask_for_password ssh_command );

my $result_dir;                  # absolute path to the results. It is something like 
                                 # "/var/opt/lsb/test/manager/results/$result_subdir"
my $result_subdir;               # the name of the directory with results. 
my $output_file_name;            # -o <filename>
my $certification_mode = 0;      # --cert
my $manual_tests_present = 1;    # --no-man-tests (inverse)
my $automatic_tests_present = 1; # --no-auto-tests (inverse)
my $send_mail = 0;               # --send-mail=somebody@somewhere
my $tester_email = "";
my $rlocation = "";              # --ssh "user@host:/dir"
my $ssh_password = "";             # --ssh-password

my %parsed_files = ();

# HTML colors used in the report.
my $dark_gray = "rgb(64, 64, 64)";
my $gray = "rgb(200, 200, 200)";
my $neutral_gray = "rgb(230, 230, 230)";
my $bright_gray = "rgb(240, 240, 240)";
my $fluorescent_green = "rgb(64, 255, 64)";
my $fluorescent_yellow = "rgb(255, 255, 64)";
my $fluorescent_red = "rgb(255, 96, 96)";
my $pale_blue = "rgb(220, 220, 240)";
my $dark_blue = "rgb(0, 0, 128)";
my $dark_red = "rgb(128, 0, 0)";
my $red = "rgb(255, 0, 0)";
my $green = "rgb(0, 160, 0)";
my $blue = "rgb(0, 0, 255)";

# Statistics for each test
my %stats_failures_reported = ();
my %stats_warnings_reported = ();
my %stats_fips_reported = ();
my %stats_notices_reported = ();
my %stats_tests_total = ();
my %stats_tests_passed = ();
my %stats_tests_skipped = ();
my %stats_tests_quelled = ();
my %stats_tests_expected = ();
my %stats_tests_skips_unexpected = ();

# Summary statistics
my $tests_total = 0;
my $tests_passed = 0;
my $tests_quelled = 0;
my $tests_skipped = 0;

my $tests_cancelled = 0;
my $failures_reported = 0;
my $overall_verdict = "";
my $auto_verdict = 0;
my $man_verdict = 0;

# Manual tests stasistics
my $man_tests_total = 0;
my $man_tests_passed = 0;
my $man_tests_quelled = 0;
my $man_tests_skipped = 0;
my $man_failures_reported = 0;
my $man_tests_cancelled = 0;

# Trial tests stasistics
my $trial_tests_total = 0;
my $trial_tests_passed = 0;
my $trial_tests_quelled = 0;
my $trial_tests_skipped = 0;
my $trial_failures_reported = 0;
my $trial_tests_cancelled = 0;


my $start_time;
my $stop_time;

# Host information
my $host_machine;
my $host_name;
my $host_platform;
my $host_os;
my $host_kernel;
my $etc_issue;
my $lsb_version;
my $package_versions;
my $global_start_time;
my $global_finish_time;

my $test_comments;
my $tester_name;
my $organization;

# The problem database, hashed by the "test_name test_point" strings.
my %problem_database;

# The final exit code, which is 0 if all tests have passed, or 1 if some
# failures were detected.
my $exit_code = 0;

# The report file.
my $report;

# The single-line summary file.
my $summary_line;

# This variable is set to 1 if part of the problem table was already printed
# (at least the header), false if not (this implies that no problems were
# found so far).
my $problem_table_printed = 0;

# A monotonously increasing, unique identifier of a message box.
my $msgbox_id = 1;

# Copy of the input line that was processed last.
my $line = "";

# File with console output log.
my $console_log_file = 'test_logs/tests_output.8888';

#---------------------------------------------------------------------

# The journal file being processed.
my $journal_file;
my $journal_name;

# The title of the journal being processed.
my $journal_title = "";
my @journal_titles = ();

my %journal_group = ();

my %is_trial = ();

# The HTML anchors used for marking this journal in the resulting report.
my %journal_anchors = ();

# Reports for each test suite
my %reports = ();  # $reports{$journal_title} = $html_report_text

sub report {
    $reports{$journal_title} .= shift;
}

my %test_versions = (); # $test_versions{$journal_title} = <test_version>
#---------------------------------------------------------------------

# Map of TET result codes to human-readable severity names.
my %result_code_map = (
    0 => "Passed",
    1 => "Failed",
    2 => "Unresolved",
    3 => "Unused",
    4 => "Unsupported",
    5 => "Untested",
    6 => "Uninitiated",
    7 => "Unreported",
    65 =>  "Time expired",
    101 => "Warning",
    102 => "FIP",
    103 => "Unimplemented",
    104 => "Unapproved",
);

#---------------------------------------------------------------------
my $END_LEFT_  = "left end";
my $END_RIGHT_ = "right end";
my $END_NO_    = "no end";

sub html_entities {
    my $text = shift;
    my $end_ = shift;
    
    if (!defined $end_)
    {
        $text =~ s/(^[\n\r\s]+|[\n\r\s]+$)//g;
    }
    else
    {
        if ($end_ eq $END_LEFT_)
        {
            $text =~ s/(^[\n\r\s]+)//g;
        }
        elsif ($end_ eq $END_RIGHT_)
        {
            $text =~ s/([\n\r\s]+$)//g;
        }
    }
    # Escape HTML entities
    $text =~ s/<br \/>/\n/gi;
    $text =~ s/&/&amp;/g;
    $text =~ s/</&lt;/g;
    $text =~ s/>/&gt;/g;
    $text =~ s/[\n\r]+/<br \/>\n/g;
    
    return $text;
}
#---------------------------------------------------------------------
sub fail {
    my $text = shift;
    
    print STDERR "$text\n";
    exit(-1);
}

#---------------------------------------------------------------------
# Prints time and date in a human-readable format
sub format_time {
    my $timestamp = shift;
    
    my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday) = 
        localtime ($timestamp);

    return sprintf ("%02d:%02d:%02d %02d-%02d-%04d\n", $hour, $min, $sec,
        $mday, ($mon + 1), ($year + 1900) );
}

#---------------------------------------------------------------------
# Loads the problem database which is used for automatic identification
# of known test suite failures, warnings and false positives.
sub load_problem_database {
    my $file_name = shift;

    # Open the database file.
    unless ( open(FILE, $file_name) ) {
        print "Could not open problem database $file_name\n";
        return;
    }

    # The problem record being read.
    my $problem;

    while ( my $line = <FILE> ) {
        
        $line =~ s/^\s+//; # Remove spaces in the beginning of the line
        $line =~ s/\s+$//; # Remove trailing spaces
        
        next if $line eq "";   # Skip empty lines
        next if $line =~ /^#/; # Skip comments

        if ( !$problem ) {
            # The 'test' line starts a new test record.
            # It contains the name of the test and the test point number.
            if ($line =~ /^test\s+(.*)\s+(\d+)\s*$/) {
                my $test_name = $1;
                my $test_point = $2;

                # Clear the old record and prepare a clean one.
                $problem = {
                    'test_name' => $test_name,
                    'test_point' => $test_point,
                };
                
                next;
            }
            else {
                fail "problem_db: $.:"
                    ." Unexpected text beyound a test block: '$line'";
            }
        }
        
        # The 'expected_severity' line indicates the severity
        # expected to be reported by the test suite.
        if ( $line =~ /^expected_severity\s+(.*)/ ) {
            my $expected_severity = lc($1);

            $problem->{'expected_severity'} = $expected_severity;
        }
        # The 'marker' line contains a marker which can be found
        # to identify the particular problem from the messages.
        # There may be (and typically are) more markers, each on its own
        # 'marker' line.
        elsif ( $line =~ /^marker\s+(.*)/ ) {
            my $marker = $1;
            if ( defined $problem->{'markers'} ) {
                push @{$problem->{'markers'}}, $marker;
            } else {
                $problem->{'markers'} = [$marker];
            }
        }
        # The 'resolution' line contains the resolution of the problem.
        elsif ($line =~ /^resolution\s+(.*)/) {
            my $resolution = $1;
            $problem->{'resolution'} = $resolution;
        }
        # If the 'quell' line is present, the problem is considered harmless
        # regardless of severity. This is used for already solved bugs
        # and false positives.
        elsif ($line =~ /^quell(\s+(.*))?/) {
            my $force_severity = $2;
            $force_severity = 1 if !$force_severity;
            
            $problem->{'quell'} = $force_severity;
        }
        # The 'comment' line holds a comment for human reader
        # which is copied to the problem report. There can be more
        # comment lines.
        elsif ($line =~ /^comment\s+(.*)/) {
            my $text = $1;
            
            $problem->{'comment'} .= "\n" if $problem->{'comment'};
            $problem->{'comment'} .= $text;
        }
        # The 'lsb_report_id' contains the LSB problem report ID.
        elsif ($line =~ /^lsb_report_id\s+(\S+)\s*$/) {
            $problem->{'lsb_report_id'} = $1;
        }
        # The 'lsb_resolution_id' contains the LSB resolution ID.
        elsif ($line =~ /^lsb_resolution_id\s+(\S+)\s*$/) {
            $problem->{'lsb_resolution_id'} = $1;
        }
        # 'dup' means that problem description should be copied for
        # another test case. 
        # NOTE: 'dup' should go AFTER all records to be copied.
        # In other words, if you put something below 'dup', it won't be copied.
        elsif ($line =~ /^dup\s+(.*)\s+(\d+)\s*$/) {
            my $dup_test_name = $1;
            my $dup_test_point = $2;

            my $copy = { %$problem };
            
            $copy->{'test_name'} = $dup_test_name;
            $copy->{'test_point'} = $dup_test_point;

            # Add the same problem record to the database with another key.
            my $record_key = $dup_test_name." ".$dup_test_point;
            
            # Add the problem record to the database.            
            if ( defined $problem_database{$record_key} ) {
                push @{$problem_database{$record_key}}, $copy;
            } else {
                $problem_database{$record_key} = [$copy];
            }
        }
        # The 'end' line finishes one problem entry.
        elsif ($line =~ /^end/) {

            # Check if this problem report has all required fields.
            unless ( $problem->{'expected_severity'} ) {
                fail "problem_db: $.:"
                    .($problem ? " ".$problem->{'test_name'}
                                ." ".$problem->{'test_point'}.":" : "")
                    ." expected_severity field is required.";
            }
            
            # The problem record database is indexed by the test name
            # and point number, joined together to a key.
            my $record_key = $problem->{'test_name'}." ".$problem->{'test_point'};
            
            # Add the problem record to the database.
            if ( defined $problem_database{$record_key} ) {
                push @{$problem_database{$record_key}}, $problem;
            } else {
                $problem_database{$record_key} = [$problem];
            }
            
            $problem = undef; # Close the problem record.
        }
        # Anything else is a syntax error.
        else {
            fail "problem_db: $.:"
                .($problem ? " ".$problem->{'test_name'}
                            ." ".$problem->{'test_point'}.":" : "")
                ." Unrecognized line: '$line'";
        }
    }
    close FILE;
    
    return 1; # OK
}
#---------------------------------------------------------------------

# Finds a problem record in the database.
#
# Arguments:
# $1 - test name
# $2 - test point
# $3 - problem severity
# $4 - test messages
sub find_problem_record
{
    my ($test_name, $test_point, $severity, $test_messages) = @_;

    # Some automatically generated problem reports have no test name or point;
    # these are not meant to be searched in the database but directly reported
    # to the user.
    unless ($test_name) {
        # Do not bother searching, return nothing immediately.
        return undef;
    }

    # Get all records matching the key
    ( %problem_database ) or die "Problem database isn't loaded.";
    
    # The basic record key is the name of the test and the test point number.
    my $record_key = $test_name . " " . $test_point;
    
    my $found_records = $problem_database{$record_key};
    
    return undef if !$found_records || !@$found_records; # Nothing found;
    
    # Try finding the problem record.
    PROBLEM_RECORD:
    foreach my $problem_record ( @$found_records ) {

        # Check if the severity matches.
        my $expected_severity = $problem_record->{'expected_severity'};
        next PROBLEM_RECORD  unless lc($severity) eq lc($expected_severity);
            
        # Look which distinguishing markers the problem should have.
        my $markers = $problem_record->{'markers'};
        
        # Check if the markers are found in the test messages.
        foreach my $marker ( @$markers ) {
            unless ( index($test_messages, $marker) >= 0 ) {
                # The message doesn't match to marker.
                next PROBLEM_RECORD; # Try another problem record.
            }
        }
        
        # All the markes match.
        # Yup, this is the problem report we looked for.
        return $problem_record;
    }

    # Nothing was found.
    return undef;
}

#---------------------------------------------------------------------

# Processes the result of a test point.
#
# Arguments:
# $1 - test number (0 if the tests are not numbered)
# $2 - test name
# $3 - test point (0 if the tests are not divided into test points)
# $4 - severity
# $5 - test messages
sub process_test_point_result {
    my ($test_number, $test_point, $test_name, $severity, $test_messages, $comment) = @_;

    $comment = "" if !defined $comment;
    
    # Sanity check: Test name must be valid.
    unless ($test_name) {
        print ("Empty test name\n");
        return;
    }

    # Compare severity strings in lowercase.
    my $lowercase_severity = lc($severity);

    # Increment the total test count.
    $stats_tests_total{$journal_title}++;

    # Passed and unused tests are not printed, only counted.
    if ($lowercase_severity eq "passed") {
        $stats_tests_passed{$journal_title}++;
        return;
    }
    elsif (($lowercase_severity eq "unused") || ($lowercase_severity eq "untested") ||
        ($lowercase_severity eq "unsupported") || ($lowercase_severity eq "unimplemented")) {
        $stats_tests_skipped{$journal_title}++;
        return;
    }

    my $quelled = 0;
    my $resolution = "";
    my $lsb_report_id = "";
    my $lsb_resolution_id = "";

    # Compare the problem description with the database.
    my $problem_record = find_problem_record($test_name, $test_point, $severity, $test_messages);
    if ($problem_record) {

        # Load the data from the problem record.
        $quelled = $problem_record->{'quell'};
        $resolution = $problem_record->{'resolution'};
        $comment .= "<br/>\n-----<br/>\n" if ( $comment ne "" );
        $comment .= $problem_record->{'comment'};
        $lsb_report_id = $problem_record->{'lsb_report_id'};
        $lsb_resolution_id = $problem_record->{'lsb_resolution_id'};

        # Add the "(quelled)" keyword to the severity if the problem is quelled.
        # This means it can be safely ignored.
        if ($quelled) {
            if ( $lsb_report_id ) {
                $severity .= " (waived)";
            } else {
                $severity .= " (quelled)";
            }
        }
    }
    
    # Print the problem to the table.
    print_problem_table_line(
        $test_number,
        $test_point,
        $test_name,
        $severity,
        $resolution,
        $comment,
        $lsb_report_id,
        $lsb_resolution_id,
        $test_messages
    );
}
#---------------------------------------------------------------------

# Prints a header of the problem table. It is called automatically
# from report_problem() when the first problem is reported.
# (If there are no problems during the whole test run, the table
# is not printed at all.)
sub print_problem_table_header() {

    report("<h4>Problem Summary</h4>\n\n");
    report("<p>Click on lines in the table to see the details " .
        "about each problem.</p>\n\n");
    report("<table width=\"100%\">\n");
    report("<thead>\n");
    report("<tr style=\"font-weight:bold; background-color:$gray;\">\n");
    report("<td>Test Name</td><td>Severity</td><td>Resolution</td>\n");
    report("</tr>\n");
    report("</thead>\n");
    report("<tbody>\n");
}

#---------------------------------------------------------------------

# Prints a line of the problem table.
#
# Arguments:
#
# $1 - test number (0 if the tests are not numbered, or if this problem line
#      is not connected to any particular test).
# $2 - test point number (0 if there is no test point).
# $3 - test name (empty if this problem line is not connected to any
#      particular test).
# $4 - severity.
# $5 - resolution (empty if unknown).
# $6 - comment
# $7 - report ID in the LSB database (empty if not yet reported).
# $8 - resolution ID in the LSB database (empty if not yet resolved).
# $9 - messages printed by the testing program.
sub print_problem_table_line
{
    my $test_number = shift;
    my $test_point = shift;
    my $test_name = shift;
    my $severity = shift;
    my $resolution = shift;
    my $comment = shift;
    my $lsb_report_id = shift;
    my $lsb_resolution_id = shift;
    my $test_messages = shift;

    # Compare the severity string case-insensitively (we are only humans).
    my $lowercase_severity = lc($severity);

    # Choose the color for the specified message severity.
    my $color;
    if ($lowercase_severity =~ /(quelled|waived)/) {
        # Anything what is quelled is harmless, so use neutral gray
        # for all quelled problems.
        $color = "gray";
        
        # Increment the quelled tests count statistic.
        $stats_tests_quelled{$journal_title}++;
    }
    elsif (($lowercase_severity eq "failed") || ($lowercase_severity eq "failure") ||
        ($lowercase_severity eq "unresolved") || ($lowercase_severity eq "unreported") ||
        ($lowercase_severity eq "uninitiated") || ($lowercase_severity eq "test suite error") || 
        ($lowercase_severity eq "time expired")
    ) {
        # Use fluorescent red for various failures.
        $color = "red";

        # Increment the number of failures reported.
        $stats_failures_reported{$journal_title}++;

        # Even a single failure in a single test is enough for flipping
        # the exit code from 0 to 1.
        $exit_code = 1;
    }
    elsif ($lowercase_severity eq "warning") {
        # Use fluorescent yellow for warnings.
        $color = "yellow";

        # Increment the number of warnings reported.
        $stats_warnings_reported{$journal_title}++;
    }
    elsif ($lowercase_severity eq "fip") {
        # Use fluorescent green for FIPs.
        $color = "green";
        
        $severity .= "<span class=\"view_srv\">&nbsp;&nbsp;<a href=\"tests_help.pl#FIP\" title=\"Help\"><img src=\"images/question.png\" width=\"16\" height=\"17\" alt=\"[?]\" /></a></span>";
        
        # Increment the number of FIPs reported.
        $stats_fips_reported{$journal_title}++;
    }
    elsif ($lowercase_severity eq "notice") {
        # Use neutral gray for harmless messages.
        $color = "gray";

        # Increment the number of notices reported.
        $stats_notices_reported{$journal_title}++;
    }
    else {
        print ("Error: Unrecognized severity code '$severity'\n");
    }
    
    # If the problem table header was not yet printed, print it now.
    unless ($problem_table_printed) {
        print_problem_table_header();
        $problem_table_printed = 1;
    }

    # Give the message box a unique ID.
    my $message_box_id = "msgbox" . $msgbox_id; $msgbox_id++;

    # Start the line (with attached script that opens the detail block
    # on clicking).
    report("<tr class=\"ptl_$color\" title='Click here!' onclick=\"flip('$message_box_id');\">");

    # Print the test name and test point (if available).
    report("<td>");
    if ($test_name) {
        report("$test_name ");
    }
    else {
        report("(none) ");
    }
    if ($test_point > 0) {
        report("<span style=\"color:$dark_blue;\">$test_point</span>\n");
    }
    report("</td>\n");

    # Print severity and resolution.
    report("<td>$severity</td>\n");
    report("<td>$resolution</td>\n");

    # Finish the line.
    report("</tr>\n");

    # Print the detail block, which is initially hidden and unrolls when
    # the table row is clicked (via the javascript above).
    report("<tr id=\"$message_box_id\" style=\"display:none;\">\n");
    report("<td colspan=\"3\">");
    report("<div style='padding-left:32px;'>");
    report("Messages from the test\n");
    report("<div style=\"color:$dark_gray;padding-left:32px;\">\n");
    if ($test_messages) {
	#
	# Parsing {noreplace} tags
	#
	report("<p style='font-family:monospace'>");	

	my $NR_TAG_BEGIN_ = "((?i)\{noreplace\})";
	my $NR_TAG_END_ =   "((?i)\{\/noreplace\})";
	
	my $tm_ = $test_messages;	
	my $num_ = 0;
	my $temp_;
	while ($tm_ =~ /$NR_TAG_BEGIN_.*?$NR_TAG_END_/)
	{	    
	    $tm_ = $';
	    if (!$num_)
	    {
		report(html_entities($`, $END_LEFT_));
	    }
	    else
	    {
		report(html_entities($`, $END_NO_));
	    }
	    $temp_ = $&;
	    $temp_ =~ s/$NR_TAG_BEGIN_|$NR_TAG_END_//g;
	    report($temp_);
	    $num_++;
	}
	report(html_entities($tm_, $END_RIGHT_));

	report("</p>");
    } else {
        report("(none)");
    }
    report("\n</div>\n");
    if ( defined $comment && $comment ne "" ) {
        report("Comment:\n");
        report("<div style=\"color:$dark_gray;padding-left:32px;\">\n");
        report("$comment\n</div>\n");
    }
    if ($lsb_report_id or $lsb_resolution_id) {
        report("LSB Report Status:\n");
        report("<div style=\"color:$dark_gray;padding-left:32px;\">\n");
        if ($lsb_report_id) {
            report("Reported to the LSB Committee as ");
            report("<a href=\"https://www.linux-foundation.org/cert/pr.php?view=$lsb_report_id\">");
            report("PR $lsb_report_id</a>.<br />\n");
        }
        if ($lsb_resolution_id) {
            report("Resolved (resolution ID is ");
            report("<span style=\"color:$dark_blue;\">$lsb_resolution_id</span>).<br />\n");
        }
        report("</div>");
    }
    report("</div>");
    if ( $color eq "red" ) {
         my $esc_journal_title = $journal_title;
         $esc_journal_title =~ s/ /%20/g;
         my $esc_test_case = $test_name.' '.$test_point;
         $esc_test_case =~ s/ /%20/g;
         report("<div align=\"right\">"
            ."<input type=\"button\" value=\"Report Bug in the Test\""
            ." onclick=\"javascript:bug_report('$result_dir/$output_file_name', '$esc_journal_title', '$esc_test_case');\""
            ." title=\"Press this button if you believe the problem is in the test, not in your system.\" /></div>"
            );
    }
    report("</td>");
    report("</tr>\n");
}

#---------------------------------------------------------------------

# Prints the footer of the problem table. Does nothing if the problem table
# was not printed at all.
sub finish_problem_table() {

    # Complain if we have not seen any tests at all; this may mean the test suite
    # terminated somehow at the very start.
    if ($stats_tests_total{$journal_title} == 0) {
        print_problem_table_line(
            0,
            0,
            "(none)",
            "Failure",
            "No tests found in the journal",
            "",
            "",
            $line
        );
    }

    # Close the table or report that no problems were detected
    if ($problem_table_printed) {
        report("</tbody>\n");
        report("</table>\n");
    }
    else {
        report("<h4>Problem Summary</h4>\n");
        report("<p><font color='#006600'><b>No problems were detected.</b></font></p>\n");
    }
}

#---------------------------------------------------------------------

# Starts a two-column table.
sub start_two_column_table {
    report("<table>\n");
    report("<tbody>\n");
}

#---------------------------------------------------------------------

# Prints a single line of a two-column table.
#
# Arguments:
# $1 - The first column.
# $2 - The second column.
# $3 (optional) - The foreground color of the second column.
# $4 (optional) - The background color of the second column.
sub two_column_table_line {
    my $first_column = shift;
    my $second_column = shift;
    my $foreground_color = shift;
    my $background_color = shift;
    
    # Build the style specifier for the second column.
    my $second_column_style = "padding: 2px; margin: 2px;";
    
    # If the second column contains a number, print it right-aligned.
#    if ($second_column =~ /^[0-9\.]*$/) {
        $second_column_style .= "text-align: right; ";
#    }
    
    # Apply the foreground and background colors if specified.
    if ($foreground_color) {
        $second_column_style .= "color: $foreground_color; ";
    }
    if ($background_color) {
        $second_column_style .= "background-color: $background_color; ";
    }
    else {
        $second_column_style .= "background-color: $neutral_gray; ";
    }
    
    # Print the line with appropriate decorations.
    return "<tr style=\"padding: 4px; margin: 4px;\">".
    "<td style=\"background-color:$gray;\">$first_column</td>\n".
    "<td style=\"$second_column_style\">$second_column</td>\n".
    "</tr>\n";
}
#---------------------------------------------------------------------

# Prints a two-column table line to the test report
sub print_two_column_table_line {
    report(two_column_table_line(@_));
}
#---------------------------------------------------------------------

# Finishes a two-column table.
sub finish_two_column_table {
    report("</tbody>\n");
    report("</table>\n");
}

#---------------------------------------------------------------------

sub print_statistics_line {
    my ($title, $value, $warn_if_nonzero, $warning_color) = @_;

    # Use neutral gray background by default.
    my $background_color = $neutral_gray;

    # If a warning is requested, change the color
    # when the value is nonzero.
    if ($warn_if_nonzero && ($value > 0) ) {
        print_two_column_table_line($title, $value, $dark_blue,
            $warning_color);
    }
    else {
        print_two_column_table_line($title, $value, $dark_blue);
    }
}


#---------------------------------------------------------------------

# Prints the statistics.
sub print_statistics() {

    report("<h4>Statistics</h4>\n");

    # Calculate the time interval between the tests start and stop.
    my $test_duration = "";

    if ( defined($start_time) && $start_time && defined($stop_time) && $stop_time )
    {
        $test_duration = $stop_time - $start_time;
        my $test_duration_sec = $test_duration % 60;
        $test_duration = int($test_duration / 60);
        my $test_duration_min = $test_duration % 60;
        my $test_duration_hour = int($test_duration / 60);
        
        if($test_duration_hour > 0) {
            $test_duration = "${test_duration_hour}h:${test_duration_min}m:${test_duration_sec}s";
        }
        else {
            $test_duration = "${test_duration_min}m:${test_duration_sec}s";
        }
    }
    
    # Print the statistics table.
    my $background_color;
    start_two_column_table();
    if ( $test_duration ) {
        print_statistics_line("Test time", $test_duration, 0);
    } else {
        if ( defined($start_time) && $start_time ) {
            print_statistics_line("Start time", format_time($start_time), 0);
        }
        if ( defined($stop_time) && $stop_time ) {
            print_statistics_line("Stop time", format_time($stop_time), 0);
        }
    }
    print_statistics_line("Total tests", $stats_tests_total{$journal_title}, 0);
    if ($stats_tests_expected{$journal_title} > 0) {
        print_statistics_line("Total tests expected", $stats_tests_expected{$journal_title}, 
            $stats_tests_total{$journal_title}<$stats_tests_expected{$journal_title}?1:0, $fluorescent_yellow );
    }
    print_statistics_line("Passed tests", $stats_tests_passed{$journal_title}, 0);
    if ($stats_tests_skipped{$journal_title} > 0) {
        print_statistics_line("Tests not relevant/not applicable", $stats_tests_skipped{$journal_title}, 0);
    }
    if ($stats_tests_skips_unexpected{$journal_title} > 0) {
        print_statistics_line("Test skips unexpected", $stats_tests_skips_unexpected{$journal_title}, 1, $fluorescent_yellow);
    }
    if ($stats_tests_quelled{$journal_title} > 0) {
        print_statistics_line("Quelled problems", $stats_tests_quelled{$journal_title}, 0);
    }
    if ($stats_fips_reported{$journal_title} > 0) {
        print_statistics_line("FIPs", $stats_fips_reported{$journal_title}, 1, $fluorescent_green);
    }
    if ($stats_warnings_reported{$journal_title} > 0) {
        print_statistics_line("Warnings", $stats_warnings_reported{$journal_title}, 1, $fluorescent_yellow);
    }
    if ($stats_failures_reported{$journal_title} > 0) {
        print_statistics_line("Failures", $stats_failures_reported{$journal_title}, 1, $fluorescent_red);
    }
    
    # Give the final verdict.
    if ($stats_failures_reported{$journal_title} > 0) {
        print_two_column_table_line("Verdict", "Failure");
    }
    elsif ($stats_warnings_reported{$journal_title} > 0) {
        print_two_column_table_line("Verdict", "Warnings");
    }
    else {
        print_two_column_table_line("Verdict", "Success");
    }

    # Finish the table.
    finish_two_column_table();

    # Sanity check: Check if the number of tests matches the sum of 
    # skipped, passed, quelled and FIPed tests.
    if (  
        ( $stats_tests_passed{$journal_title}
        + $stats_tests_skipped{$journal_title}
        + $stats_failures_reported{$journal_title}
        + $stats_warnings_reported{$journal_title}
        + $stats_fips_reported{$journal_title}
        #+ $stats_notices_reported{$journal_title} # Notices are not real journal entries
        + $stats_tests_quelled{$journal_title} )
        
        != $stats_tests_total{$journal_title} 
        )
    {
        # Oops.
        #print ("Warning. Stats inconsistensy!\n");
        #print (<<HEREDOC);
            #Passed: $stats_tests_passed{$journal_title}
            #Skipped: $stats_tests_skipped{$journal_title}
            #Failures: $stats_failures_reported{$journal_title}
            #Warnings: $stats_warnings_reported{$journal_title}
            #FIPs: $stats_fips_reported{$journal_title}
            #Quelled: $stats_tests_quelled{$journal_title}
        
            #!= Total: $stats_tests_total{$journal_title} 
#HEREDOC
    }
}

#---------------------------------------------------------------------

sub print_test_summary_line
{
    my ($title, $anchor) = @_;

    my $title_comment = sprintf ("%-25s", "<!-- $title -->");

    # Add the result into the single-line summary.
    $summary_line->print ("    $title_comment <td>");
    if ($stats_failures_reported{$journal_title} > 0) {
        $summary_line->print ("<a href=\"test_log.html#$anchor\">");
        $summary_line->print ("<img src=\"failure_icon.png\" alt=\"F\">");
        $summary_line->print ("</a>");
    }
    elsif ($stats_warnings_reported{$journal_title} > 0) {
        $summary_line->print ("<a href=\"test_log.html#$anchor\">");
        $summary_line->print ("<img src=\"warning_icon.png\" alt=\"W\">");
        $summary_line->print ("</a>");
    }
    elsif ($stats_tests_total{$journal_title} > 0) {
        $summary_line->print ("<a href=\"test_log.html#$anchor\">");
        $summary_line->print ("<img src=\"ok_icon.png\" alt=\"OK\">");
        $summary_line->print ("</a>");
    }
    else {
        $summary_line->print ("<a href=\"test_log.html#$anchor\">");
        $summary_line->print ("<img src=\"failure_icon.png\" alt=\"Un\"></a>");
    }
    $summary_line->print ("</td>\n");
}

#---------------------------------------------------------------------

sub parse_journal
{
    $journal_name = shift;
    my $descr_filename;
    if ( $journal_name =~ /\.INFO$/ ) {
        $descr_filename = $journal_name;
        $journal_name =~ s/\.INFO$//;
    } else {
        $descr_filename = $journal_name.".INFO";
    }
    my $title = "";
    my $anchor = "";
    my $description = "";
    my $test_version = "";
    $start_time = undef;
    $stop_time = undef;
    my $trial = '';
    my $failed = undef;

    # Parse INFO file
    my $res;
    if ( !-f $descr_filename || !open(FILE, "$descr_filename") ) {
        print ("Couldn't open file '$descr_filename' for reading.\n");
        $anchor = $journal_name;
        $anchor =~ s/.*\///;
        $anchor =~ s/\..*$//;
        $title = $anchor;
    } else {
        while ( my $line = <FILE> ) {
            if ( $line =~ /^\s*TITLE:\s+(.*)$/ ) { $title = $1; }
            elsif ( $line =~ /^\s*ANCHOR:\s+(.*)$/ ) { $anchor = $1; }
            elsif ( $line =~ /^\s*JOURNAL_FILENAME:\s+(.*)$/ ) { 
                $journal_name = $1;
                my $path = $descr_filename;
                $path =~ s/\/[^\/]*//;
                $journal_name = "$path/$journal_name";
            }
            elsif ( $line =~ /^\s*VERSION:\s+(.*)$/ ) { $test_version = $1; }
            elsif ( $line =~ /^\s*START_TIME:\s+(.*)$/ ) { $start_time = $1; }
            elsif ( $line =~ /^\s*STOP_TIME:\s+(.*)$/ ) { $stop_time = $1; }
            elsif ( $line =~ /^\s*DESCRIPTION:\s+(.*)$/ ) { $description .= $1; }
            elsif ( $line =~ /^\s*TRIAL:\s+(.*)$/ ) { $trial = $1; }
            elsif ( $line =~ /^\s*FAILED:\s+(.*)$/ ) { $failed = $1; }
            elsif ( $line !~ /^\s*#/ && $line !~ /^\s*$/ ) {
                $description .= $line;
            }
            else {
                print "Warning: Syntax error in '$descr_filename' at line '$line'.";
            }
        }
        $parsed_files{$descr_filename} = 1;
        close FILE;
    }
    
    unless ($journal_name) {
        print "Syntax error in $descr_filename"
             ." (JOURNAL_FILENAME: $journal_name)\n";
        return 1;
    }
    
    $journal_title = $title;
    if ( $journal_title =~ /^([^\\\/]+)[\\\/]([^\\\/]+)[\\\/]([^\\\/]+)$/ ) {
        $journal_title = $3;
        $journal_group{$journal_title}{TOP_GROUP} = $1;
        $journal_group{$journal_title}{SUB_GROUP} = $2;
    } else {
        $journal_group{$journal_title}{TOP_GROUP} = "Unsorted";
        $journal_group{$journal_title}{SUB_GROUP} = "";
    }
    
    # Store the journal title and anchor.
    push @journal_titles, $journal_title;
    $journal_anchors{$journal_title} = $anchor;
    $test_versions{$journal_title} = $test_version;
    $is_trial{$journal_title} = $trial;
    
    $problem_table_printed = 0;

    $stats_failures_reported{$journal_title} = 0;
    $stats_warnings_reported{$journal_title} = 0;
    $stats_fips_reported{$journal_title} = 0;
    $stats_notices_reported{$journal_title} = 0;
    $stats_tests_total{$journal_title} = 0;
    $stats_tests_passed{$journal_title} = 0;
    $stats_tests_skipped{$journal_title} = 0;
    $stats_tests_quelled{$journal_title} = 0;
    $stats_tests_expected{$journal_title} = 0;
    $stats_tests_skips_unexpected{$journal_title} = 0;
    
    # If journal isn't closed then close it.
    if ( $journal_file ) { close_journal (); }
    
    # Open the journal file.
    print ("Reading journal: $journal_name\n");
    $journal_file = new IO::File;
    unless (open ($journal_file, $journal_name)) {
        if ($journal_group{$journal_title}{SUB_GROUP} =~ m/\bManual\b/) {
            # Journal for manual test is not found - manual tests are just not finished
            $man_tests_cancelled++;
            report( "<p><font color='#A00000'><b>No result.</b></font>"
                ." The manual test <b>$journal_title</b> was not executed yet.</p>"
            );
            print "Failed to read journal: '$journal_name'.\n"
                ."The manual test $journal_title was not executed yet.\n";
        }
        else {
            # The journal file could not be open; mention this in the log
            # (so that we know that something was missing) and return.
            
            unless ( $certification_mode && $is_trial{$journal_title} ) {
                $tests_cancelled++;
            }
            
            report( '<p>' );
            if ( defined($failed) ) {
                report( '<font color="#A00000"><b>'.$failed.'</b></font> ' );
            } else {
                report( "<font color='#A00000'><b>No result.</b></font> "
                    ." $journal_title failed to run, interrupted"
                    ." or crashed during the execution. "
                );
                print "Failed to read journal: '$journal_name'."
                    ."\n$journal_title failed to run, interrupted or crashed"
                    ." during the execution.\n($journal_name was not present).\n";
            }
            report ( 'Please see the '
                    .'<a target="_blank"'
                            .' href="/get.pl'.$result_dir.'/'.$console_log_file.'"'
                            .' class="view_srv">log file</a>'
                    .'<a target="_blank" href="'.$console_log_file.'"'
                            .' class="view_as_is">log file</a>'
                    .' for details.'
                ) if (-f "$result_dir/$console_log_file");
            report( '</p>' );
            
            print_test_summary_line ($journal_title, $anchor);
        }
        $journal_file = undef;
    } else {
        $parsed_files{$journal_name} = 1;
        if ( $description ) {
            report("<p>".html_entities($description)."</p>");
        }
        report("<p><a href=\"/get.pl$result_dir/$journal_name\" class=\"view_srv\" target=\"_blank\">Open the journal file</a>"
              ."<a href=\"$journal_name\" class=\"view_as_is\" target=\"_blank\">Open the journal file</a></p>" );
    }
    
    if ( $journal_name =~ /\.journal$/ ) {
        parse_tet_journal();
    }elsif ( $journal_name =~ /\.tcljournal$/ ) {
        parse_tcl_journal();
    }elsif ( $journal_name =~ /\.pythonjournal$/ ) {
        parse_python_journal();
    }elsif ( $journal_name =~ /\.simplejournal$/ ) {
        parse_simple_journal();
    }elsif ( $journal_name =~ /\.raptorjournal$/ ) {
        parse_raptor_journal();
    }
}

#---------------------------------------------------------------------

sub close_journal
{
    unless ($journal_file) { return; }
    
    # Finish the problem table.
    finish_problem_table();

    # Print the statistics.
    print_statistics();

    # Add the result into the single-line summary.
    print_test_summary_line ( $journal_title, $journal_anchors{$journal_title} );

    # Close the journal file.
    $journal_file->close ();

    # Reset state variables.
    $journal_file = undef;
    $journal_title = "";
}

#---------------------------------------------------------------------

# The main TET journal parser function which loads the TET journal
# and produces a HTML report. Uses stdin if the input file name is "-".
#
# Parameters:
#    $1 - the title to be printed before the report (typically the name
#         of the test the journal belongs to).
#    $2 - a HTML anchor to produce (useful for referencing to the results
#         of the test in a larger document).
#    $3 - the file name of the journal.
sub parse_tet_journal
{
    unless ($journal_file) { return; }
    
    # This flag is set to nonzero if we are inside a test point.
    my $inside_test = 0;
    my $inside_test_point = 0;

    # Configuration variables (from lines with code 30)
    my %config_vars = ();

    # Read the journal line by line.
    my ($test_name, $test_number, $test_point, $test_messages, $test_comment);
    my $test_points_in_this_case;
    my $line_number = 0;
    while ($line = <$journal_file>) {
        $line_number++;
        # Each journal line consists of three parts, separated by
        # a vertical bar:
        #
        # 1. A numeric line code, determining the meaning of the line.
        # 2. Additional data, format depends on line code.
        # 3. Additional text (message for the user, or some comment).
        # e.g. '400|0 1 1 20:04:00|IC Start'
        unless ($line =~ /^(\d*)\|([^\|]*)\|(.*)$/) {

            # Lines that do not conform to this format are reported
            # to the user and skipped.
            print_problem_table_line(
                0,      # test number
                0,      # test point
                "(none)",       # test name
                "Warning",      # severity
                "Unrecognized entry in test journal (wrong syntax)",       # resolution
                "$journal_name($line_number)",      # comment
                "",     # LSB problem report id
                "",     # resolution id
                $line   # test messages
            );
            next;
        }
        my $line_code = $1;
        my $line_data = $2;
        my $line_text = $3;
        
        # Determine the type of the line and process it appropriately.
        if ($line_code == 0) {
            # Code 0 -- Test Case Controller Start.
            # The line structure is:
            # 0|version time date|who
            unless ($line_data =~ /^(\S*)\s*(\S*)\s(\S*)$/) {
                print_problem_table_line(
                    0,      # test number
                    0,      # test point
                    "(none)",       # test name
                    "Warning",      # severity
                    "Syntax error in test journal",     # resolution
                    "$journal_name($line_number)",      # comment
                    "",     # LSB problem report id
                    "",     # resolution id
                    $line   # test messages
                ); 
                next;
            }

            # This line should be one of the first;
            # if it comes in the middle of normal results, it is strange.
            if ($problem_table_printed) {
                print_problem_table_line(
                    0,
                    0,
                    "(none)",
                    "Warning",
                    "Test Case Start entry appears in the middle of the journal",
                    "$journal_name($line_number)",
                    "",
                    "",
                    "$line_data\n$line_text"
                );
            }
            else {
                my $test_engine_version = $1;
                my $test_start_time = $2;
                my $test_start_date = $3;
                my $test_command_line = $line_text;

                # Print the test engine version, time and date to the output.
                report("<h4>Test suite information</h4>");
                start_two_column_table();
                print_two_column_table_line("Test&nbsp;Engine&nbsp;Version", $test_engine_version);
                print_two_column_table_line("Test&nbsp;Started&nbsp;At", "$test_start_time $test_start_date");
                print_two_column_table_line("Test&nbsp;Command&nbsp;Line", "$test_command_line");
                finish_two_column_table();
            }
        }
        elsif ($line_code == 5) {

            # Code 5 -- Local System Information.
            # The line structure is:
            # 5|sysname nodename release #date machine|text
            unless ($line =~ /^5\|(\S*)\s*(\S*)\s*(\S*)\s*#(.*)\s(\S+)\|/) {
                print_problem_table_line(
                    0,
                    0,
                    "(none)",
                    "Warning",
                    "Syntax error in test journal",
                    "$journal_name($line_number)",
                    "",
                    "",
                    $line);
                next;
            }

            # The system information should be one of the first items;
            # if it comes in the middle of normal results, it is strange.
            if ($problem_table_printed) {
                print_problem_table_line(
                    0,
                    0,
                    "(none)",
                    "Warning",
                    "System info entry appears in the middle of the journal",
                    "$journal_name($line_number)",
                    "",
                    "",
                    "$line_data\n$line_text"
                );
            }
            else {
                my $system_name = $1;
                my $node_name = $2;
                my $release = $3;
                my $date = $4;
                my $machine = $5;

                # Print the local system data.
                report("<h4>Tested system information</h4>\n");
                start_two_column_table();
                print_two_column_table_line("System Name", $system_name);
                print_two_column_table_line("Node Name", $node_name);
                print_two_column_table_line("Release", $release);
                print_two_column_table_line("Release Date", $date);
                print_two_column_table_line("Machine", $machine);
                finish_two_column_table();
            }
        }
        elsif ($line_code == 10) {

            # Code 10 -- Test Case Start.
            # The line structure is:
            # 10|activitynum testname hh:mm:ss|list of invocable components
            # * Timestamp is omitted in perl test reports. (Bug?)
            unless ($line_data =~ /^(\d+)\s+(.*)\s+\S+$/) {
                print_problem_table_line(
                    0,
                    0,
                    "(none)",
                    "Warning",
                    "Syntax error in test journal",
                    "$journal_name($line_number)",
                    "",
                    "",
                    $line);
                next;
            }
            
            my $activitynum = $1;
            my $testname_param = $2;
            my $timestamp = $4 if defined $4;

            if ($test_name) {
                # The previous test case was not properly finished;
                # issue a warning.
                0 and # quelled
                print_problem_table_line(
                    $activitynum,
                    0,
                    $test_name,
                    "Warning",
                    "Test case ($testname_param) starts before "
                        ."the previous one ($test_name) has finished",
                    "$journal_name($line_number)",
                    "",
                    "",
                    ""
                );
            }

            # Store the test name and number.
            $test_number = $activitynum;
            $test_name = $testname_param;

            # Reset the test point number and counter.
            $test_point = 0;
            $test_points_in_this_case = 0;
        }
        elsif ($line_code == 15) {
            # Code 15 -- Test Case Manager Start.
            # Not interesting to us.
        }
        elsif ($line_code == 20) {
            # Code 20 -- Local/Remote System Configuration Start.
            # Not interesting to us.
        }
        elsif ($line_code == 30) {

            # Code 30 -- Configuration Variables.
            # The line structure is:
            # 30||VARIABLE=value
            if ($line =~ /^30\|\|([^=]+)=(.*)/) {
                my $varname = $1;
                my $value = $2;
                
                $config_vars{$varname} = $value;
            }
        }
        elsif ($line_code == 40) {
            # Code 40 -- Configuration End.
            
            # Print the local system data.
            report("<h4>Configuration information</h4>\n");
            start_two_column_table();
            for my $varname qw(VSX_NAME LSB_VERSION LSB_MODULES TET_VERSION XT_FONTPATH) {
                if ( exists $config_vars{$varname} ) {
                    print_two_column_table_line($varname, $config_vars{$varname});
                }
            }
            finish_two_column_table();
        }
        elsif ($line_code == 50) {
            # Code 50 -- Test Case Controller Message.
            print_problem_table_line(
                $test_number,
                $test_point,
                $test_name,
                "Test Suite Error",
                "Test case controller reports a problem",
                "",
                "",
                "",
                $line_text
            );
        }
        elsif ($line_code == 70) {
            # Code 70 -- Scenario Information.

            # LSB-runtime-test only:
            # for the .os sections need to double expected counts to allow
            # for the macro versions.
                if ( $line =~ /^70\|.*total tests in (ANSI|POSIX|LSB|PTHR).os ([0-9]+)/ ) {
                    $stats_tests_expected{$journal_title} += $2;
                    $stats_tests_expected{$journal_title} += $2;
                }
                elsif ( $line =~ /^70\|.*total tests in ([A-Za-z.0-9_-]+) ([0-9]+)/ ) {
                    $stats_tests_expected{$journal_title} += $2;
                }
                elsif ( $line =~ /total\s+tests\s+in/ ) {
                    complain ("$journal_name($line_number): A suspicious record in journal.")
                }
        }
        elsif ($line_code == 80) {
            # Code 80 -- Test Case End.
            unless ($test_name) {
                # Test case end found when no test case was started;
                # issue a warning.
                print_problem_table_line(
                    0,
                    0,
                    "(none)",
                    "Warning",
                    "Test Case End entry found when no test case seems to be running",
                    "$journal_name($line_number)",
                    "",
                    "",
                    $test_messages,
                    $test_comment
                );
            }

            # Warn if the test case seems to be empty.
            if ($test_points_in_this_case == 0) {
                print_problem_table_line(
                    $test_number,
                    0,
                    $test_name,
                    "Warning",
                    "Test case contains no test points",
                    "",
                    "",
                    "",
                    ""
                );
            }

            # Clear the test name, number, and messages.
            $test_name = "";
            $test_number = 0;
            $test_messages = "";
            $test_comment = "";
            $test_points_in_this_case = 0;
        }
        elsif ($line_code == 90) {
            # Code 90 -- User Abort.
            print_problem_table_line(
                0,
                0,
                "(none)",
                "Test Suite Error",
                "Test suite execution interrupted by the user",
                "",
                "",
                "",
                $line_text
            );
        }
        elsif ($line_code == 100) {
            # Code 100 -- Captured Output.
            # Not interesting to us.
        }
        elsif ($line_code == 110) {
            # Code 110 -- Build Start.
            # Not interesting to us.
        }
        elsif ($line_code == 130) {
            # Code 130 -- Build End.
            # Not interesting to us.
        }
        elsif ($line_code == 200) {
            # Code 200 -- Test Purpose Start.
            # The structure of the line is as follows:
            # 200|activitynum pointnum hh:mm:ss|comment
            unless ($line =~ /^\d*\|(\d*)\s*(\d*)/) {
                print_problem_table_line(
                    0,
                    0,
                    "(none)",
                    "Warning",
                    "Syntax error in test journal",
                    "$journal_name($line_number)",
                    "",
                    "",
                    $line);
                next;
            }

            # Check if the previous test point was properly finished.
            if ($test_point > 0) {
                if ($test_point == $2) {
                    # Some test suites have a strange habit of re-starting
                    # the same test point more than once.
                    print_problem_table_line(
                        $test_number,
                        $test_point,
                        $test_name,
                        "Notice",
                        "Test was restarted in the middle",
                        "$journal_name($line_number)",
                        "",
                        "",
                        ""
                    );
                }
                else {
                    # Notify about a disappeared test.
                    print_problem_table_line(
                        $test_number,
                        $2,
                        $test_name,
                        "Warning",
                        "Test point starts before the previous one ($test_point) has finished",
                        "$journal_name($line_number)",
                        "",
                        "",
                        ""
                    );
                }
            }
            
            # Store the test point number.
            $test_point = $2;
            
            # Clear the messages.
            $test_messages = "";
            
            # Line text part
            if ( defined $line_text && $line_text ne "" && $line_text !~ /^T[CP] Start$/ ) {
            	$test_comment = "Test Purpose Comment: ".$line_text;
            } else {
                $test_comment = "";
            }

            # Increment the number of test points.
            $test_points_in_this_case++;
        }
        elsif ($line_code == 220) {
            # Code 220 -- Test Purpose Result.
            # The structure of the line is as follows:
            # 220|activitynum pointnum result time|comment
            unless ($line_data =~ /^(\d+)\s+(\d+)?\s+(\d+)\s+(\S+)$/) {
                complain("Syntax error in journal");
                next;
            }
            my $activitynum = $1;
            my $pointnum = $2 if defined $2;
            my $result_code = $3;
            my $timestamp = $4;

            # Print a warning if the test purpose seems to never been started.
            if ($test_point == 0) {
                print_problem_table_line(
                    $test_number,  $test_point,  $test_name,
                    "Warning",
                    "Test result entry found but the test point had not started",
                    "$journal_name($line_number)",
                    "",  "",  $test_messages, $test_comment
                );
            }

            # Print a warning if the test purpose number is different from
            # what we expected.
            if ( defined $pointnum && $test_point != $pointnum) {
                print_problem_table_line(
                    $test_number,  $test_point,  $test_name,
                    "Warning",
                    "Test purpose number at the end ($2) is different " .
                        "from the starting one ($test_point)",
                    "$journal_name($line_number)",
                    "",  "",  $test_messages, $test_comment
                );
            }

            # Find the meaning of the result code (severity).
            if ( !defined $result_code_map{$result_code} ) {
                print_problem_table_line(
                    $test_number,  $test_point,  $test_name,
                    "Failure",
                    "Unrecognized test result code ($result_code)",
                    "$journal_name($line_number)",
                    "",  "",  $test_messages, $test_comment
                );
                next;
            }

            my $severity = $result_code_map{$result_code};

            # Process the result.
            process_test_point_result(
                $test_number,
                $test_point,
                $test_name,
                $severity,
                $test_messages,
                $test_comment,
            );

            # Clear the test point number.
            $test_point = 0;
        }
        elsif ($line_code == 310) {
            # 310 -- Clean Start.
            # Not interesting to us.
        }
        elsif ($line_code == 320) {
            # 320 -- Clean End.
            # Not interesting to us.
        }
        elsif ($line_code == 400) {
            # 400 -- Invocable Component Start.
            # Not interesting to us.
        }
        elsif ($line_code == 410) {
            # 410 -- Invocable Component End.
            # Not interesting to us.
        }
        elsif ($line_code == 510) {
            # 510 -- Error reported by the underlying test suite.
            print_problem_table_line(
                $test_number,
                $test_point,
                $test_name,
                "Test Suite Error",
                "Error reported by the test suite",
                "$journal_name($line_number)",
                "",
                "",
                $line_text
            );
        }
        elsif ($line_code == 520) {
            # 520 -- Message from the underlying test suite.

            # Coalesce consequent rows of spaces in the message
            # to a single space. This improves matching with markers.
            my $message = $line_text;
            $message =~ s/\s+/ /g;

            # Add the message to the accumulated messages.
            $test_messages .= $message."\n";
        }
        elsif ($line_code == 600) {
            # 600 -- Parallel Start.
            # Not interesting to us.
        }
        elsif ($line_code == 620) {
            # 620 -- Parallel End.
            # Not interesting to us.
        }
        elsif ($line_code == 630) {

            # 630 -- Implied Sequential Start.
            # Not interesting to us.
        }
        elsif ($line_code == 640) {
            # 640 -- Implied Sequential End.
            # Not interesting to us.
        }
        elsif ($line_code == 700) {
            # 700 -- Repeat Start.
            # Not interesting to us.
        }
        elsif ($line_code == 720) {
            # 720 -- Repeat End.
            # Not interesting to us.
        }
        elsif ($line_code == 730) {
            # 730 -- Timed Loop Start.
            # Not interesting to us.
        }
        elsif ($line_code == 740) {
            # 740 -- Timed Loop End.
            # Not interesting to us.
        }
        elsif ($line_code == 750) {
            # 750 -- Random Start.
            # Not interesting to us.
        }
        elsif ($line_code == 760) {
            # 760 -- Random End.
            # Not interesting to us.
        }
        elsif ($line_code == 800) {
            # 800 -- Remote Start.
            # Not interesting to us.
        }
        elsif ($line_code == 820) {
            # 820 -- Remote End.
            # Not interesting to us.
        }
        elsif ($line_code == 830) {
            # 830 -- Distributed Start.
            # Not interesting to us.
        }
        elsif ($line_code == 840) {
            # 840 -- Distributed End.
            # Not interesting to us.
        }
        elsif ($line_code == 900) {
            # 900 -- Test Case Controller End.
            # Not interesting to us.
        }
        else {
            # Other lines are notified to the user and otherwise ignored.
            print_problem_table_line(
                $test_number,
                $test_point,
                $test_name,
                "Notice",
                "Unrecognized journal entry (code $line_code)",
                "$journal_name($line_number)",
                "",
                $line
            );
        }
    }

    # Finish the journal processing and close it.
    close_journal ();
}

#---------------------------------------------------------------------

# The Tcl-style journal parser function.
sub parse_tcl_journal {
    
    unless ($journal_file) { return; }

    # Read the journal line by line.
    my $severity = "";
    my $test_number = 1;
    my ($test_name, $test_messages);
    while ($line = <$journal_file>) {

        # A test record message ends with a line with a rough format
        # '==== testname result'
        if ($line =~ /^=+\s*([a-zA-Z0-9_\-\.]+)\s*FAILED/) {

            unless ($test_name) {

                # Oops, we have missed something...
                print_problem_table_line(
                    $test_number,
                    0,
                    $1,
                    "Failure",
                    "End of test record found but no start was seen",
                    "",
                    "",
                    "",
                    $line
                );
            }

            # Process the result.
            process_test_point_result(
                $test_number,
                0,
                $test_name,
                "Failed",
                $test_messages
            );

            # Clear the previous test messages.
            $test_messages = "";

            # Maintain an artificial sequence of test numbers.
            $test_number++;

            # Finish the test record.
            $test_name = "";
        }

        # A failure report starts with a line with a rough format
        # '==== testname errormessage result'.
        elsif ($line =~ /^=+\s*([a-zA-Z0-9_\-\.]+)\s*(.+)\s*FAILED/) {

            # Load the test name.
            $test_name = $1;

            # Add the error message to the test messages.
            $test_messages .= $2."\n";
        }

        # A line that contains only the name of the test (which ends
        # with '.test', marks a passed test.
        elsif ($line =~ /^([a-zA-Z0-9_\-]+)\.test/) {
            process_test_point_result(
                $test_number,
                0,
                $1,
                "Passed",
                ""
            );

            # Maintain an artificial sequence of test numbers.
            $test_number++;
        }

        # Anything other?
        else {
            if ($test_name) {

                # We are inside a test point result record, so it will be
                # a test message.
                $test_messages .= $line."\n";
            }
            else {

                # Anything other is better ignored (there is too much noise
                # for considering unrecognized lines to be errors).
            }
        }
    }

    close_journal ();
}

#---------------------------------------------------------------------

# The Python-style journal parser function.
sub parse_python_journal
{
    unless ($journal_file) { return; }

    # Read the journal line by line.
    my $severity = "";
    my $inside_test_point = 0;
    my $test_number = 1;
    my ($test_name, $test_point, $test_messages);
    while ($line = <$journal_file>) {

        # Are we reading a test point record?
        if ($inside_test_point) {

            # Any line that resembles a start of another test record
            # stops this test record.
            if ($line =~ /^test_/) {
                $inside_test_point = 0;

                # Process this result.
                process_test_point_result(
                    $test_number,
                    $test_point,
                    $test_name,
                    "Failure",
                    $test_messages
                );

                # Clear the previous test messages.
                $test_messages = "";
            }
            else {

                # Just append this line to the test record messages
                # and continue by the next line.
                $test_messages .= $line."\n";
                next;
            }
        }

        # Any other line is supposed to end the test point record.
        $inside_test_point = 0;

        # A line that contains only the name of the test (a single
        # identifier), marks a passed test.
        if ($line =~ /^([a-zA-Z0-9_]+)\s*$/) {
            process_test_point_result(
                $test_number,
                $test_point,
                $1,
                "Passed",
                ""
            );

            # Tests are not naturally numbered so figure out a monotonous sequence.
            $test_number++;
        }

        # A line in form 'testname skipped -- ...' marks a test
        # that was skipped, and the end of the line contains
        # the description why.
        elsif ($line =~ /^([a-zA-Z0-9_]+)\s*skipped\s*--\s*(.*)$/) {
            process_test_point_result(
                $test_number,
                $test_point,
                $1,
                "Unsupported",
                ""
            );

            # Tests are not naturally numbered so figure out a monotonous sequence.
            $test_number++;
        }

        # A line in form 'test testname failed -- ...' marks a failed test.
        elsif ($line =~ /^test\s*([a-zA-Z0-9_]+)\s*(failed|crashed)\s*--\s*(.*)$/) {

            # Get the test name.
            $test_name = $1;
            $test_point = 1;

            # We are reading the test failure report.
            $inside_test_point = 1;

            # Get the first line of the messages.
            $test_messages = $3;
        }
        
        elsif ($line =~ /^(\d+)\s+skips\s+unexpected/ ) {
            my $skips_unexpected = $1;
            $stats_tests_skips_unexpected{$journal_title} = $skips_unexpected;
        }
        
        # Other lines are silently ignored.
    }

    close_journal ();
}

#---------------------------------------------------------------------

sub parse_simple_journal
{
    unless ($journal_file) { return; }

    # Read the journal line by line.
    my $severity = "";
    my $inside_test_point = 0;
    my $test_number = 1;
    my ($test_name, $test_point, $test_messages, $test_result);
    while ($line = <$journal_file>) {

        # Each line in form 'test name: result' should contain
        # the result of one test.
        if ($line =~ /^\s*([^:]*)\s*:\s*(\w*)/) {

            # Tests are not naturally numbered so figure out a monotonous sequence.
            $test_number++;

            # Load the name and result.
            $test_name = $1;
            $test_result = $2;

            # An optional detail message can follow.
            if ($line =~ /^\s*([^:]*)\s*:\s*(\w*)\s*:\s*(.*)$/) {
                $test_messages = $3;
            }

            # Check for failures.
            if ($test_result eq "failed") {
                process_test_point_result(
                    $test_number,
                    0,
                    $test_name,
                    "failed",
                    $test_messages
                );
            }
            elsif ($test_result eq "warning") {
                process_test_point_result(
                    $test_number,
                    0,
                    $test_name,
                    "warning",
                    $test_messages
                );
            }
            elsif ($test_result eq "OK") {
                process_test_point_result(
                    $test_number,
                    0,
                    $test_name,
                    "passed",
                    $test_messages
                );
            }
            else {

                # Anything other is an error in the journal.
                print_problem_table_line(
                    $test_number,
                    0,
                    $test_name,
                    "failed",
                    "Syntax error in journal",
                    "",
                    "",
                    "",
                    $line
                );
            }
        }
        else {
            # Anything other is an error in the journal.
            print_problem_table_line(
                $test_number,
                0,
                $test_name,
                "failed",
                "Syntax error in journal",
                "",
                "",
                "",
                $line
            );
        }
    }

    close_journal ();
}
#---------------------------------------------------------------------

# The Raptor-style journal parser function 
#
sub parse_raptor_journal
{
    unless ($journal_file) { return; }

    # Read the journal line by line.
    my $severity = "";
    my $test_number = 1;
    my ($test_name, $test_messages);
    while ($line = <$journal_file>) {
        
        chomp($line);
        
        if ($severity eq "FAILED") {
            if ($line !~ /^Checking\s/) {
                $test_messages .= $line."\n";
            } else {
                # End of previous ERROR message
                $test_messages = "Here is a diff between what expected and "
                    ."what obtained:\n".$test_messages;
                # Process the result.
                process_test_point_result(
                    $test_number,
                    0,
                    $test_name,
                    "FAILED",
                    $test_messages
                )
            }
        }
        
        # The journal starts with line
        # Testing legal rdf/xml
        if ($line =~ /^Testing /) {
            next;
        }

        # A test case looks like
        # Checking $test_name (ok|FAILED)
        if ($line =~ /^Checking\s+([^\s]+)\s+(ok|FAILED)$/)
        {
            $test_name = $1;
            my $result = $2;
            
            $test_messages = "";
            if ( $result eq "ok" ) {
                process_test_point_result(
                    $test_number,
                    0,
                    $test_name,
                    "Passed",
                    ""
                );
                $severity = "";
            }
            elsif ( $result eq "FAILED" ) {
                $severity = "FAILED";
            }
            
            $test_number++;
        }
    }

    close_journal ();
}
#---------------------------------------------------------------------

sub report_index_line {
    my $journal_title = shift;
    my $display_journal_title = shift;
    
    # Cut the left spaces to bring it out of the hyperlink
    my $leftspace = "";
    if ($display_journal_title =~ /^((\s|&nbsp;)+)(.*)/ ) {
        $leftspace = $1;
        $display_journal_title = $3;
    }
  
    my $text = "";
    # Row starts
    $text .= "<tr style='background-color:".$neutral_gray."'>";
    
    # The first column: the hyperlink to the anchor in the report
    $text .= "<td>";
    $text .= "&nbsp;&nbsp;&nbsp;&nbsp;".$leftspace
            ."<b><a href='#".$journal_anchors{$journal_title}."'"
            ." title='Click here to proceed to details'>"
            .$display_journal_title."</a></b>";
    $text .= "</td>\n";
    
    # The second column: version of the testsuite
    $text .= "<td>";
    if ($journal_title !~ /::/ && $test_versions{$journal_title} ) {
        $text .= "&nbsp; <font size='-1' color='#555555'>v "
            .$test_versions{$journal_title}."</font>";
    }
    $text .= "</td>\n";
    
    # The 3rd column: status
    my $bg_color = $neutral_gray;
    my $status = "";
    if ($stats_failures_reported{$journal_title} > 0) {
        $bg_color = $fluorescent_red;
        $status = "Failures: ".$stats_failures_reported{$journal_title};
    }
    elsif ($stats_warnings_reported{$journal_title} > 0) {
        $bg_color = $fluorescent_yellow;
        $status = "Warnings: ".$stats_warnings_reported{$journal_title};
    }
    elsif ($stats_tests_passed{$journal_title} == 0) {
        if ( $journal_group{$journal_title}{SUB_GROUP} =~ m/\bManual\b/) {
            $bg_color = $fluorescent_yellow;
            $status = "<span title=\"'No Result' means the manual test"
            ." was not executed yet.\">No Result"
            ."<span class=\"view_srv\">&nbsp;&nbsp;<a href=\"tests_help.pl#No_result\" title=\"Help\"><img src=\"images/question.png\" width=\"16\" height=\"17\" alt=\"[?]\" /></a></span></span>";
        } else {
            $bg_color = '#FF8090';
            $status = "<span title=\"'No Result' means the tests failed"
            ." to run, interrupted or crashed during the execution.\">No Result"
            ."<span class=\"view_srv\">&nbsp;&nbsp;<a href=\"tests_help.pl#No_result\" title=\"Help\"><img src=\"images/question.png\" width=\"16\" height=\"17\" alt=\"[?]\" /></a></span></span>";
        }
    }
    elsif ($stats_tests_expected{$journal_title} > 0 
          && $stats_tests_expected{$journal_title} > $stats_tests_total{$journal_title} ) 
    {
        $bg_color = $fluorescent_yellow;
        $status = "Warning";
    }
    else {
        $bg_color = $fluorescent_green;
        $status = "Success";
    }
    # still the 3rd column: status
    $text .= "<td style=\"background-color:$bg_color\">$status";
    $text .= "</td>\n";
   
    # The 4th column: number of passed tests
    $text .= "<td>";
    $text .= "<span style='color:black'>Passed: ".
            $stats_tests_passed{$journal_title};
    $text .= "</span>";
    $text .= "</td>\n</tr>\n";
    return $text;
}

sub compile_report {
    # Print Title
    $report->print ("<h1>LSB automatic test results for $host_os on $host_platform</h1>\n");
    if (-f "$result_dir/$console_log_file") {
        $report->print ("The complete console output log is available "
            ."<a target=\"_blank\" href=\"/get.pl$result_dir/$console_log_file\" class=\"view_srv\">here</a>"
            ."<a target=\"_blank\" href=\"$console_log_file\" class=\"view_as_is\">here</a>.<br/><br/>");
    }
    $report->print ("<h3>Test Info</h3>\n");
    
    # Print general information
    $report->print (
    "<table>\n"
    .two_column_table_line("Tests Started At", $global_start_time )
    .two_column_table_line("Tests Finished At", $global_finish_time )
    .two_column_table_line("Host Name", $host_name)
    .two_column_table_line("Host OS", $host_os)
    .two_column_table_line("Host Machine", $host_machine)
    .two_column_table_line("Host Platform", $host_platform)
    .two_column_table_line("Host Kernel", $host_kernel)
#    .two_column_table_line("/etc/issue", "<pre>$host_etc_issue</pre>")
    ."</table>\n"
    );
    
    # Build index
    my %groups = ();
    foreach my $journal_title( @journal_titles ) {
        my $top_group = $journal_group{$journal_title}{TOP_GROUP};
        my $sub_group = $journal_group{$journal_title}{SUB_GROUP};
        if ( exists $groups{$top_group}{$sub_group} ) {
            push @{$groups{$top_group}{$sub_group}}, $journal_title;
        } else {
            my @ar = ($journal_title);
            $groups{$top_group}{$sub_group} = \@ar;
        }
    }
    
    my @journal_titles_sorted = ();
    my %subclasses = ();
    
    # Print the index
    $report->print("<h3>Tests executed</h3>");
    $report->print("<table border='0' cellpadding='2' cellspacing='1'>");
    
    foreach my $top_group ( "Certification Tests", "Trial Use Tests (Optional)", 
                            "Unofficial Tests", "Unsorted"
    ) {
        if ( exists $groups{$top_group} ) {
            $report->print( "<tr><td colspan='4' style='background-color:#b0b0ff'>"
                           ."&nbsp;<b>$top_group</b></td></tr>");
            foreach my $sub_group ("Command Tests", "Static Interface Tests", ""
                , "Runtime Interface Tests", "Automated Application Battery Tests",
                "Manual Application Battery Tests"
            ) {
                if ( exists $groups{$top_group}{$sub_group}) {
                    my @sub_elems = @{$groups{$top_group}{$sub_group}};
                    $report->print( 
                         "<tr><td colspan='4' style='background-color:#d0d0f0'>"
                        ."&nbsp;&nbsp;&nbsp;<b>$sub_group</b></td></tr>" );
                    foreach my $journal_title (sort @sub_elems) {
                        my $display_journal_title = $journal_title;
                        if ( $journal_title =~ /(.*)::(.*)/ ) {
                            my $subclass = $1;
                            $display_journal_title = "&nbsp;&nbsp;&nbsp;&nbsp;".$2;
                            if ( ! exists $subclasses{$subclass} ) {
                                $report->print( "<tr style='background-color:#dbdbe8'><td>"
                                                ."&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<font color='#224466'><b>$subclass Tests:</b></font></td>"
                                                ."<td>&nbsp; <font size='-1' color='#555555'>"
                                                .($test_versions{$journal_title}?"v "
                                                .$test_versions{$journal_title}:"")."</font></td>"
                                                ."<td colspan='2'></td></tr>"
                                               );
                                $subclasses{$subclass} = 1;
                            }
                        }
                        $report->print( report_index_line($journal_title, $display_journal_title) );
                        push @journal_titles_sorted, $journal_title;
                    }
                }
            }
        }
    }
    $report->print("</table>");
    
   
    # Test journals
    foreach $journal_title( @journal_titles_sorted ) {
        my $test_report = ($reports{$journal_title} or "");
        # Print the title.
        $report->print("<hr />\n");
        $report->print("<h2><a name=\"".$journal_anchors{$journal_title}
                ."\"></a><font color=\"#336699\">".$journal_title."</font> &nbsp; <font size='-1' color='#888888'>"
                .($test_versions{$journal_title}?"v ".$test_versions{$journal_title}:"")."</font></h2>\n");
        
        $report->print ( $test_report );
    }
    
    # Print the host info to the one-line summary.
    $summary_line->print ("<table>\n <tr>\n");
    $summary_line->print ("    <!-- OS version -->       <td class=\"title\">$host_os</td>\n");
    $summary_line->print ("    <!-- Architecture -->     <td class=\"title\">$host_machine</td>\n");
    $summary_line->print (" </tr>\n</table>\n");
}

#---------------------------------------------------------------------

sub sum {
    my $sum_main = 0;
    my $sum_manual = 0;
    my $sum_trial = 0;
    my $arr = $_[0];
    foreach (keys %$arr) {
        if ( $certification_mode && $is_trial{$_} ) {
            $sum_trial += $arr->{$_};
        }
        elsif ( $certification_mode 
                && $journal_group{$_}{TOP_GROUP} =~ m/\bUnsorted\b/
        ) {
            # Unsorted tests are considered as trial too.
            $sum_trial += $arr->{$_};
        }
        elsif ($journal_group{$_}{SUB_GROUP} =~ m/\bManual\b/) {
            $sum_manual += $arr->{$_};
        }
        else {
            $sum_main += $arr->{$_};
        }
    }
    return ($sum_main, $sum_manual, $sum_trial);
}

#---------------------------------------------------------------------

sub generate_INFO_file {
    my $INFO_file_name = "INFO";
    
    my $file;
    unless (open($file, "> $INFO_file_name")) {
        print "Couldn't open file '$INFO_file_name' for writing.\n";
        return;
    }

    print $file ""
        .($tests_total + $man_tests_total + $trial_tests_total)." Tests Total\n"
        .($tests_passed + $man_tests_passed + $trial_tests_passed)." Tests Passed\n"
        .($failures_reported + $man_failures_reported)." Failures\n"
        .($trial_failures_reported ? $trial_failures_reported." Failures in trial test\n" : '' )
        .($tests_quelled + $man_tests_quelled + $trial_tests_quelled)." Tests Quelled\n"
        .($tests_skipped + $man_tests_skipped + $trial_tests_skipped)." Tests Skipped\n"
        ;
    
    if ( $tests_cancelled > 0 ) {
        print $file "There are cancelled tests\n";
    }
    
    print $file "Verdict: $overall_verdict\n";

    if ($certification_mode) {
        print $file "Certification\n";
    }    
    if ($man_verdict == -1) {
        print $file "In progress\n";
    }
    if (!$automatic_tests_present) {
        print $file "No automatic\n";
    }
    
    print $file "\n".$test_comments;
    
    close($file);
}

#---------------------------------------------------------------------

sub Init {
    # Load the problem database (it should be in the same directory
    # as the main program, so simply get the program name and replace
    # the file part by 'problem_db').
    
    my $database_file_name = "$script_directory/problem_db";
    load_problem_database($database_file_name);
    
    # Open the report files.
    $report = new IO::File;
    unless ($report->open ("> $result_dir/test_log.html")) {
        fail ("Could not open the report file for writing");
    }
    $report->print ("<html>\n");
    $report->print ("
<head>
<style type=\"text/css\">
.ptl_gray { background: #E6E6E6;}
.ptl_gray:hover { background: #D6D6F0;}
.ptl_green { background: #CDFFCD;}
.ptl_green:hover { background: #A0FFA0;}
.ptl_yellow { background: #FFFFCD;}
.ptl_yellow:hover { background: #FFFF90;}
.ptl_red { background: #FFCDCD;}
.ptl_red:hover { background: #FFA0A0;}
.ptl_gray,.ptl_green,.ptl_yellow,.ptl_red { cursor: pointer; }
.view_srv { display: none; }
</style>
</head>
    ");
    $report->print ("<body>\n");
    $report->print ("
<script language=\"javascript\" type=\"text/javascript\">
// <![CDATA[
function flip (message_box_id) {
    msgbox = document.getElementById(message_box_id);
    if (msgbox.style.display != 'none') {
        msgbox.style.display = 'none';
    }
    else {
        msgbox.style.display = '';
    }
}
function bug_report(journal_path, test_suite, test_case) {
    if (confirm('A new E-mail to lf_lsbcert\@lists.linux-foundation.org will now be created.\\nPlease, specify all the necessary details and attach the journal files:\\n\\n' +
        journal_path + '\\n\\nContinue?'))
    {
        document.location = 'mailto:lf_lsbcert\@lists.linux-foundation.org?Subject=Problem%20report:%20' + test_suite + '%20/%20' + test_case;
    }
}
// ]]>
</script>
            ");
        
    # Open the one-line-summary file.
    $summary_line = new IO::File;
    unless ($summary_line->open ("> $result_dir/one_line_summary.html")) {
        fail ("Could not open the one-line summary file for writing"
             ." ($result_dir/one_line_summary.html)" );
    }
    $summary_line->print ("<html>\n<body>\n");
}

#---------------------------------------------------------------------

sub format_line($$) {
    my $test_res = $_[0];
    my $test_name = $_[1];
    my $res = "";
    $res .= "<td width=\"100\"".(($test_res <= -2) ? " style=\"color: #7f7f7f;\"" : "")."><b>$test_name</b></td>\n";
    if ($test_res == -3) {
        $res .= "<td style=\"text-align: center; color: #7f7f7f; background-color: #e0e0e0;\"><b>NOT REQUIRED</b></td>";
    }
    elsif ($test_res == -2) {
        $res .= "<td style=\"text-align: center; color: #7f7f7f; background-color: #e0e0e0;\"><b>NOT SELECTED</b></td>";
    }
    elsif ($test_res == -1) {
        $res .= "    <td style=\"text-align: center; color: #000000; background-color: #ffff00;\"><b><a href=\"tests_appbat.pl?test_run=$result_subdir\" style=\"color: #000000;\" title=\"Run manual tests\">NOT FINISHED YET</a></b></td>";
    }
    elsif ($test_res == 0) {
        $res .= "<td style=\"text-align: center; color: #ffffff; background-color: #ff0000;\"><b>FAILED</b></td>";
    }
    elsif ($test_res == 1) {
        $res .= "    <td style=\"width: 100px; text-align: center; color: #000000; background-color: #00ff00;\"><b>PASSED</b></td>";
    }
}

# Generates the general report with status of the automatic and manual tests.
sub generate_summary_report {
    # Open the report files.
    my $c_report = new IO::File;
    unless ($c_report->open ("> $result_dir/sum_rep.html")) {
        print ("Could not open the report file for writing\n");
        return;
    }
    
    my $email_body = "Body= Name: $tester_name;\n"
        ." Organization: $organization;\n OS: $host_os;\n"
        ." Architecture: $host_platform;\n LSB Version: $lsb_version";
    $email_body =~ s/ /%20/g;
    $email_body =~ s/[\r\n]/%0A/g;

    $c_report->print ("<html>\n<body>\n");

    my $show_date = $result_subdir;
    $show_date =~ s/^[^\-]*\-[^\-]*\-([^\-]*)\-([^\-]*)\-([^\-]*)\-([^\-]*)h\-([^\-]*)m\-([^\-]*)s/$3.$2.$1&nbsp;$4:$5:$6/;

    my $line_auto = format_line($auto_verdict, "Automatic Tests");
    my $line_man = format_line($man_verdict, "Manual Tests");

    $c_report->print (<<DATA);
<h1>Summary Report for Test Run of $show_date</h1>

<h2>Test Execution Status</h2>
<table class="main" width="350">
  <tr>
    $line_auto
    <td valign="middle" rowspan="2"><a href="tests_results.pl?details=$result_subdir">View detailed report</a></td>
  </tr>
  <tr>
    $line_man
  </tr>
</table>
DATA

    if ($man_verdict == -1) {
        $c_report->print ("<p align=\"justify\">The manual tests are not executed yet. You can <a href=\"tests_appbat.pl?test_run=$result_subdir\"><b>run them now</b></a>.</p>\n");
    }

    if ($overall_verdict eq "Failed") {
        $c_report->print (<<DATA);
<h2>Analyze the Failing Tests</h2>
<p align="justify">
Please analyze the <a href="tests_results.pl?details=$result_subdir"><b>detailed report</b></a>
to understand the reasons of each fail and classify them into the following groups:</p>
<ol>
<li><b>Confirmed FAILs</b>, which are due to the real inconsistencies of your system with the LSB standard. To continue with the LSB certification, you have to fix your implementation to remedy such FAILs and then rerun the tests again.</li>
<li><b>False FAILs</b>, which you believe are due to the incorrect tests.
Please report such FAILs to the <a href="mailto:lf_lsbcert\@lists.linux-foundation.org?Subject=False%20positive%20FAIL&amp;$email_body">lf_lsbcert\@lists.linux-foundation.org</a> list.
If the problem reported is confirmed you will be granted a <i>waiver</i> so the test failure will not affect your ability to certify.</li>
<li><b>Unknown FAILs</b>, which you have no idea about. Please ask help at the <a href="mailto:lf_lsbcert\@lists.linux-foundation.org?Subject=Unknown%20FAIL&amp;$email_body">lf_lsbcert\@lists.linux-foundation.org</a> list.</li>
</ol>
<p align="justify"><b>Note:</b> only test results with all the FAILs waived are eligible for certification.</p>
DATA
    }
    
    $c_report->print (<<DATA);
<h2>Locating Test Journals</h2>
<p align="justify">You can find the consolidated test journals of this test run at the following path (click to download to another location):<br />
<a href="get.pl$result_dir/$output_file_name">$result_dir/$output_file_name</a><br />
You need to attach them when applying for certification or when communicating with the support staff.</p>
DATA

    if ($certification_mode and ($man_verdict != -1)) {
        $c_report->print (<<DATA);
<h2>Apply for Certification</h2>
<p align="justify">With all the tests performed successfully, you may apply for certification. 

</p>
<p align="justify">Organizational details of the certification procedure are described at the <a href="http://www.linux-foundation.org/en/Certification" target="_blank">Certification Home Page</a>.</p>
DATA
# You will need to send an E-mail to <a href="mailto:lf_lsbcert\@lists.linux-foundation.org?Subject=Certification%20test%20results&amp;$email_body">lf_lsbcert\@lists.linux-foundation.org</a> with all the tests journals attached (see <i>Locating Test Journals</i> above).
    }
    
    if ($certification_mode and ($man_verdict != -1)) {
        # Good certification results
        $c_report->print (<<DATA); # DO NOT BREAK THE LINE
<span id="cert-integration" style="display:none">FILE=$result_dir/$output_file_name;CERTIFICATION=1</span>
DATA
    } else {
        # Non-applicable for certification results.
        $c_report->print (<<DATA);
<span id="cert-integration" style="display:none">FILE=$result_dir/$output_file_name</span>
DATA
    }
    
    $c_report->print (<<DATA);
<h2>Viewing this Page Again</h2>
<p align="justify">This page has been saved as a part of the test run results. You may view it at any time by clicking particular test result at the <a href="tests_results.pl">Results</a> page.</p>

</body>
</html>
DATA

    # Close the report files.
    $c_report->close ();
}

#---------------------------------------------------------------------

# Read and return file contents
sub read_file {
    my $filename = shift;
    my $text = "";
    
    unless (open FILE, $filename) {
        print ("Can't open $filename\n");
        return "";
    }
    
    while (<FILE>) {
        $text .= $_;
    }
    close FILE;
    
    chomp($text);
    return $text;
}

sub read_test_info {
    $host_machine       = read_file("$result_dir/host_info/host_machine");
    $host_name          = read_file("$result_dir/host_info/host_name");
    $host_platform      = read_file("$result_dir/host_info/host_platform");
    $host_os            = read_file("$result_dir/host_info/host_os");
    $host_kernel        = read_file("$result_dir/host_info/host_kernel");
    $etc_issue          = read_file("$result_dir/host_info/etc_issue");
    $lsb_version        = read_file("$result_dir/host_info/lsb_version");
    $package_versions   = read_file("$result_dir/host_info/package_versions");
    $etc_issue          = read_file("$result_dir/host_info/etc_issue");

    $global_start_time  = read_file("$result_dir/autotest_info/start_time");
    $global_finish_time = read_file("$result_dir/autotest_info/finish_time");
    $test_comments      = read_file("$result_dir/autotest_info/test_comments");
    $tester_name        = read_file("$result_dir/autotest_info/tester_name");
    $organization       = read_file("$result_dir/autotest_info/organization");
    
    if (open(INFO_FILE, "$result_dir/INFO")) {
        while (<INFO_FILE>) {
            s/[\r\n]//g;
            if (m/^No automatic$/) {
                $automatic_tests_present = 0;
                last;
            }
        }
        close(INFO_FILE);
    }
}
    
#---------------------------------------------------------------------

# Create INFO files for the manual tests
sub create_manual_INFOs {
    my $appbat_data_dir = "$script_directory/../webui/public_html/app_battery";
    if (opendir(APPBAT_DIR, $appbat_data_dir)) {
        foreach my $test (grep /^$lsb_version/, readdir(APPBAT_DIR)) {
            $test =~ s/^$lsb_version-lsb-(.*)\.inc/$1/;
            if (open(TEST_JOURNAL_INFO, ">$result_dir/test_journals/lsb-$test.simplejournal.INFO")) {
                print TEST_JOURNAL_INFO 
                    "TITLE: Certification Tests/Manual Application Battery Tests/$test\n"
                    ."ANCHOR: lsb_$test\n"
                    ."JOURNAL_FILENAME: lsb-$test.simplejournal\n";
                close(TEST_JOURNAL_INFO);
            }
        }
        closedir(APPBAT_DIR);
    }
}

#---------------------------------------------------------------------

# Generates a HTML report and detects test failures.
# Returns the return code of the reporting utility, which is 0 if all tests
# have passed, 1 if some tests have failed.
sub generate_report
{
    print ("Generating HTML report\n");
    
    #Go to the output directory
    chdir($result_dir);
    
    # Unpack all *.tar.gz archives (autotest results)
    opendir DIR, "." or die "Can't open directory $result_dir: $!\n";
    my @files = readdir(DIR);
    foreach my $file (@files) {
        if ( $file =~ /\.tar\.gz$/ ) {
            system ("tar -xkzf $file 2>/dev/null");
            unlink($file);
        }
    }
    
    # Read the test information.
    read_test_info();

    if ($manual_tests_present) {
        create_manual_INFOs();
    }

    $result_subdir = $result_dir;
    chomp($result_subdir);
    $result_subdir =~ s/\/+$//; # remove trailing /
    $result_subdir =~ s/^.*\///; # remove everything before the last /
    
    unless ($output_file_name) {
        $output_file_name = "lsb-autotest-$result_subdir";
    }
    if ($output_file_name !~ /\./) {
        $output_file_name .= ".tgz";
    }
    
    # Open a report file and print a header
    Init();
    
    # Parse each test journal
    opendir DIR, "test_journals" or die "Can't open directory $result_dir/test_journals: $!\n";
    @files = readdir(DIR);
    foreach my $file (@files) {
        if ( $file =~ /journal(.INFO)?$/ ) {
            if ( !$parsed_files{"test_journals/$file"} ) {
                parse_journal("test_journals/$file");
            }
        }
    }

    # Sum up statistics
    ($failures_reported, $man_failures_reported, $trial_failures_reported) 
            = sum( \%stats_failures_reported );
    ($tests_total, $man_tests_total, $trial_tests_total)
            = sum( \%stats_tests_total );
    ($tests_passed, $man_tests_passed, $trial_tests_passed)
            = sum( \%stats_tests_passed );
    ($tests_quelled, $man_tests_quelled, $trial_tests_quelled)
            = sum( \%stats_tests_quelled );
    ($tests_skipped, $man_tests_skipped, $trial_tests_skipped)
        = sum( \%stats_tests_skipped );
    
    # Give the final verdict
    if ( ($failures_reported + $man_failures_reported > 0) || ($tests_cancelled > 0) ) {
        $overall_verdict = "Failed";
    }
    else {
        $overall_verdict = "Success";
    }
    if ($automatic_tests_present) {
        $auto_verdict = ((($failures_reported > 0) || ($tests_cancelled > 0)) ? 0 : 1);
    }
    else {
        $auto_verdict = -2;
    }
    if ($manual_tests_present) {
        if ($man_tests_cancelled > 0) {
            $man_verdict = -1;
        }
        elsif ($man_failures_reported > 0) {
            $man_verdict = 0;
        }
        else {
            $man_verdict = 1;
        }
    }
    else {
        $man_verdict = ($certification_mode ? -3 : -2);
    }

    # Write INFO file
    generate_INFO_file ();
    
    compile_report ();
    
    generate_summary_report ();
    
    if ( -e "$result_dir/$output_file_name" ) {
        unlink("$result_dir/$output_file_name");
    }
    
    # Finish
    
    # Close the report files.
    $report->print ("</body>\n</html>\n");
    $report->close ();
    
    $summary_line->print ("</body>\n</html>\n");
    $summary_line->close ();
    
    return $exit_code;
}

sub send_output_by_email {
    # Send the results by e-mail if requested.
    print("Sending an e-mail\n");
    
    # Use sendmail utility
    &sendmail ( { 
        TO      => $tester_email,
        SUBJECT => "LSB Distribution Testkit results for $host_name",
        TEXT    => "Please see the attached archive for the test results.",
        ATTACHMENT => "$result_dir/$output_file_name",
                } 
        );
    return 1;
}

sub upload_results
{
    my $spawn_command = "rsync \\\"$result_dir/$output_file_name\\\""
        ." \\\"$rlocation/$output_file_name\\\"";
    my $result = ssh_command($ssh_password, $spawn_command);
    if ( 0 != $result ) {
        print "Operation failed: $spawn_command\n";
        return $result;
    }
    if ( $rlocation =~ m/^([^:]+):(.*)$/ ) {
        my $user_at_host= $1;
        my $dir_at_host= $2;
        my $command = "cd \\'$dir_at_host\\' && "
                ." tar xzf \\'$output_file_name\\' && "
                ." mv \\'$output_file_name\\' \\'$result_subdir\\' && "
                ." echo \\'$result_subdir\\' >> HISTORY";
        my $ssh_command = "ssh $user_at_host '$command || echo \\\"!E*\\\"'";
        $result = ssh_command($ssh_password, $ssh_command);
    }
    return $result;
}

sub read_command_line
{
    my $argc = scalar(@ARGV);
    if ($argc < 1) {
        print "Usage: $0 <test_result_dir>\n";
        exit 1;
    }
    $result_dir = make_path_absolute($ARGV[0]);
    
    my $output_file_name_expected = 0;
    my $rlocation_expected = 0;
    
    for (my $i = 0; $i<$argc; $i++) {
        my $arg = $ARGV[$i];
        
        if ($output_file_name_expected) {
            $output_file_name_expected = 0;
            $output_file_name = $arg;
        }
        if ( $rlocation_expected ) {
            $rlocation_expected = 0;
            $rlocation = $arg;
        }
        
        if ($arg eq "-o") {
            $output_file_name_expected = 1;
        }
        elsif ($arg eq "--cert") {
            $certification_mode = 1;
        }
        elsif ($arg eq "--no-man-tests") {
            $manual_tests_present = 0;
        }
        elsif ($arg eq "--no-auto-tests") {
            $automatic_tests_present = 0;
        }
        elsif ($arg =~ m/--send-mail=(.*)/) {
            $send_mail = 1;
            $tester_email = $1;
        }
        elsif ($arg =~ m/--upload/) {
            $rlocation_expected = 1;
        }
    }
}

### BEGIN ###

read_command_line();

if ( $rlocation ) { 
    if ( defined( $ENV{SSH_PASSWORD} ) ) {
        $ssh_password = $ENV{SSH_PASSWORD};
    } else {
        $ssh_password = ask_for_password("SSH Password: ");
    }
}
    
$exit_code = generate_report();

# make the tarball with results
chdir("$result_dir/..");
system ("tar czf $output_file_name $result_subdir && mv $output_file_name $result_subdir");

# Send the report by e-mail.
send_output_by_email() if ($send_mail);

upload_results() if ($rlocation);

exit($exit_code);

### END ###
 
