#!/usr/bin/perl -w

# LSB Test Execution Framework
# Web-server Module (dtk-server.pl)
#
# Copyright (C) 2007-2008 The Linux Foundation. All rights reserved.
#
# This program has been developed by ISP RAS for LF.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

use strict;
use Socket;
use IO::Socket;
use POSIX ":sys_wait_h";
use Fcntl ":flock";
use Cwd qw(cwd abs_path);

# $DOCUMENT_ROOT - path to directory connected to http://localhost:<port>/
# It's the 'public_html' subdirectory in the dtk-server.pl script directory.
my $SERVER_ROOT = abs_path($0);
$SERVER_ROOT =~ s!/[^/]+$!!;
my $DOCUMENT_ROOT = "$SERVER_ROOT/public_html";
my $APP_DATA = "/var/opt/lsb/test/manager";

my $port = (shift or 8888);
my $LOG_FILE = "$APP_DATA/log/dtk-server.log.8888";
my $first_log_line = "Log file refreshed. HTTP server is running on port '$port'. PID: $$.";

my @index_docs = ("index.html", "index.htm", "index.pl", "index.cgi");
my %mime_types;

sub LOG($;$) {
	my $is_empty = !(-s $LOG_FILE);
	if (open(LOGFILE, ">>$LOG_FILE")) {
		flock(LOGFILE, LOCK_EX);
		my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst) = localtime();
		my $date = sprintf("%02d.%02d.%04d %02d:%02d:%02d", $mday, ($mon + 1), ($year + 1900), $hour, $min, $sec);
		if ($is_empty and !$_[1]) {
			print LOGFILE "$date\t$first_log_line\n";
		}
		print LOGFILE "$date\t".$_[0]."\n";
		flock(LOGFILE, LOCK_UN);
		close(LOGFILE);
	}
}

sub HTTP_Responce(@) {
	my ($client, $code) = @_;
	print $client "HTTP/1.0 $code", Socket::CRLF;
	print $client Socket::CRLF;
	print $client <<DATA;
<html>
<head>
<title>$code</title>
</head>
<body>
<h1>$code</h1>
<p><br>
Please, contact the <a href="mailto:linux\@ispras.ru">ISP RAS support team</a> to report about this error.<br>
See the <a href="/get.pl$LOG_FILE" target="_blank">server log file</a> for details.</p>
</body>
</html>
DATA
	LOG("[$$]: Responce: $code");
}

sub stop_server {
	LOG("Caught terminating signal for HTTP server. Stopping...");
	exit;
}

$SIG{"INT"}  = \&stop_server;
$SIG{"TERM"} = \&stop_server;

my $server = new IO::Socket::INET(Proto => 'tcp',
				LocalPort => $port,
				Listen => SOMAXCONN,
				Reuse => 1);
die "Could not create server socket on port '$port': $!" unless $server;

unlink($LOG_FILE);
LOG("Started HTTP server on port '$port'. PID: $$.", 1);

if (open(EXT_INFO, "$SERVER_ROOT/mime.types")) {
	while (<EXT_INFO>) {
		if (/^[^\s#]/) {
			s/[\r\n]//g;
			my ($read_type, @read_exts) = split(/\s+/);
			foreach (@read_exts) {
				$mime_types{$_} = $read_type;
			}
		}
	}
	close(EXT_INFO);
}

chdir($DOCUMENT_ROOT);
while (my $client = $server->accept()) {
	$client->autoflush(1);

	my $child = fork();
	if (!defined($child)) {
		close($client);
		LOG("Connection accepted. ERROR: Could not start fork!");
		HTTP_Responce($client, "500 Internal Server Error");
	}
	elsif ($child == 0) {
		select(undef, undef, undef, 0.1);
		LOG("Connection accepted. ID: $$.");
		my %request = ();
		$request{"COOKIES"} = "";
		$request{"HOST"} = $client->sockhost().":".$client->sockport();
		{
			local $/ = Socket::CRLF;
			my $content_length = 0;

			while (<$client>) {
				chomp;
				# headers
				if (/\s*(\w+)\s*([^\s\?]+)(\?([^\s]*))?\s*HTTP\/(\d\.\d)/) {
					$request{"METHOD"} = uc($1);
					$request{"URL"} = $2;
					$request{"HTTP_VERSION"} = $5;
					$request{"GET_ARGS"} = $4 ? $4 : "";
					$request{"GET_ARGS"} =~ s/#.*$//;
					LOG("[$$]: Requested: $_");
				}
				elsif (m/^Content-Length:/i) {
					(my $name, $content_length) = split(/:\s*/, $_, 2);
				}
				elsif (m/^Cookie:/i) {
					(my $name, $request{"COOKIES"}) = split(/:\s*/, $_, 2);
				}
				elsif (m/^Host:/i) {
					(my $name, $request{"HOST"}) = split(/:\s*/, $_, 2);
				}
				elsif (/^$/) {
					if ($content_length) {
						read($client, $request{"POST_ARGS"}, $content_length);
					}
					else {
						$request{"POST_ARGS"} = "";
					}
					last;
				}
			}
			$request{"URL"} =~ s/%([0-9a-fA-F]{2})/chr(hex($1))/eg;
		}

		# process request
		if (($request{"METHOD"} eq 'GET') or ($request{"METHOD"} eq 'POST') or ($request{"METHOD"} eq 'HEAD')) {
			# Forbid '..' sequence inside URL
			if ($request{"URL"} =~ m/\.\./) {
				HTTP_Responce($client, "403 Forbidden");
			}
			else {
				if ($request{"URL"} eq "/") {
					foreach(@index_docs) {
						if (-f $DOCUMENT_ROOT."/".$_) {
							$request{"URL"} = "/".$_;
							last;
						}
					}
				}
				my $get_pl_pos = index($request{"URL"}, "/get.pl/");
				if ($get_pl_pos != -1) {
					$get_pl_pos += 7;
					$request{"GET_ARGS"} = "file=".substr($request{"URL"}, $get_pl_pos);
					$request{"URL"} = substr($request{"URL"}, 0, $get_pl_pos);
				}
				my $localfile = $DOCUMENT_ROOT.$request{"URL"};
				if (-f $localfile)
				{
					my $buffer;
					my $is_executable = ($request{"URL"} =~ m/\.(pl|cgi|php)$/);
					my $opened;
					if ($is_executable) {
						$ENV{"DTKM_GET_ARGS"} = $request{"GET_ARGS"};
						$ENV{"DTKM_POST_ARGS"} = $request{"POST_ARGS"};
						$ENV{"DTKM_COOKIES"} = $request{"COOKIES"};
						$ENV{"DTKM_HOST"} = $request{"HOST"};
						$ENV{"DTKM_PEER_IP"} = $client->peerhost();
						$ENV{"DTKM_PORT"} = $port;
						$ENV{"DTKM_DOCUMENT_ROOT"} = $DOCUMENT_ROOT;
						$ENV{"DTKM_APP_DATA"} = $APP_DATA;
						$opened = open(FILE, "$localfile 2>/tmp/dtk_script_error.$$ |");
					}
					else {
						$opened = open(FILE, $localfile);
					}
					if ($opened)
					{
						my $mime_type;
						my $first_line;
						my $null_output = 0;
						if ($is_executable) {
							$first_line = <FILE>;
							if (defined($first_line)) {
								print $client $first_line if ($client->connected);
								$first_line =~ s/[\r\n]//g;
								my $resp = $first_line;
								$first_line .= " <from script>";
								while($resp ne "") {
									$resp = <FILE>;
									print $client $resp if ($client->connected);
									$resp =~ s/[\r\n]//g;
								}
							}
							else {
								$null_output = 1;
								$first_line = "<null> <from script>";
							}
						}
						else {
							my $ext = ($request{"URL"} =~ m/^.*\.([^\.]*)$/) ? $1 : "";
							$mime_type = ($mime_types{$ext} or "text/html");
							$first_line = "HTTP/1.0 200 OK ($mime_type)";
							print $client "HTTP/1.0 200 OK", Socket::CRLF;
							print $client "Content-Type: $mime_type", Socket::CRLF;
							print $client Socket::CRLF;
						}
						while(<FILE>) {
							print $client $_ if ($client->connected and ($request{"METHOD"} ne 'HEAD'));
						}
						close(FILE);
						if ($is_executable) {
							my $err = `cat /tmp/dtk_script_error.$$`;
							if ($err) {
								LOG("[$$]: Error text received from script:\n$err");
							}
							if ($null_output and (-s "/tmp/dtk_script_error.$$" != 0)) {
								LOG("[$$]: Script execution error!");
								HTTP_Responce($client, "500 Internal Server Error");
							}
							else {
								LOG("[$$]: Responce: $first_line");
							}
						}
						else {
							LOG("[$$]: Responce: $first_line");
						}
					}
					else {
						HTTP_Responce($client, "403 Forbidden");
					}
					unlink("/tmp/dtk_script_error.$$");
				}
				else {
					HTTP_Responce($client, "404 Not Found");
				}
			}
		}
		else {
			HTTP_Responce($client, "501 Not Implemented");
		}
		LOG("[$$]: Processing finished.");
		close($client);
		close($server);
		exit;
	}
	else {
		close($client);
		my $kid;
		do {
			$kid = waitpid(-1, WNOHANG);
		} while $kid > 0;
	}
}
